/* BEGIN NOTICE

This software was written by Peter DiFalco and John Sarraille.  The
algorithm employed was inspired by the paper entitled "A Fast
Algorithm To Determine Fractal Dimensions By Box Counting", which was
written by Liebovitch and Toth, and which appeared in the journal
"Physics Letters A", volume 141, pp 386-390, (1989).

This software is not warranteed: use at your own risk.  Unlimited
permission to distribute this software is granted, so long as this
notice is included in all copies made.

Sept 2, 1991

END NOTICE */

#include <stdio.h>
#include "fd.h"
#include <math.h>


main(argc,argv)
	int	argc;
	char	*argv[];
{
extern FILE  *infile;
extern int   embed_dim,         /* How many coordinates points have. */
             *marker;           /*  To mark queues during radix sort */
extern ulong **data_array,	/* initial array of data points */
             numpoints,        /* number of data points in input file */
            *diff_test; 	 /* XOR's of pairs of coordinates */
extern long int  *next_after,    /*  array for pointers to data_array. */
                 avail;             /*  avail list pointer */
extern QHeader Q[2];       /* array of two queues for the radix sort */


   int	     i,j,m;		/* control variables */
   ulong     maxdiam=0,	        /* 2^numbits - 1: determines the scaling. */
		   /* Array to keep track of the box counts of each size.
		      Each index is the base-2 log of the corresponding box
		      size. */
             box_count_array[numbits+1];
   float     fd, fracdim();     /* Fractal Dimension */
                   /* Maximum and minimum values in input file -- we need to
		      know these in order to scale the input points. */
   double    max, min,
	   /* A buffer for inputs to stay in until they are scaled
		      and cast to integers.  */
             buf;
   long int  pointer;     /* holds indexes used as pointers to records */

   /* find out how many coordinates points have -- 1?, 2?, 3?, more? */
   if (debugging || checking) printf("getting embed_dim\n");
   embed_dim = get_e_dim(argv[1]);
   printf("Embedding Dimension taken to be: %d\n\n",embed_dim);

/*
   find max and min in input file.  This function should open the input file
   for reading, scan it to get the maximum and minimum data values it
   contains, and then seek to the beginning of the input file so that the
   code below can proceed to read the data sequentially from the beginning.
   We are having trouble getting that to work, so we have temporarily taken
   the expedient of closing the file in the function max_min, and opening it
   again below.
*/
   max_min (argv[1], &max, &min) ;
   if (debugging|| checking) printf("max is: %lf\n",max);
   if (debugging || checking) printf("min is: %lf\n",min);
     /*
        set maxdiam=2^numbits-1 -- this is the largest value that an
        unsigned integer can have with the host compiler.
        (I checked the value obtained in this manner, and it 
         is correct -- j.s.)
     */
	if (debugging || checking) printf("numbits is: %d\n",numbits);     
        for (m=0;m<numbits;m++)	maxdiam = maxdiam + (1<<m);
	if (debugging || checking) printf("maxdiam is: %lu\n",maxdiam);

	/* open input file */
	if((infile = fopen(argv[1],"r")) == NULL)
	{
		printf("Cannot open input file, %s\n",argv[1]);
		exit(-1);
	}
	 
	/* get number of data points from input file */
	fscanf(infile,"%lu",&numpoints);
	if (debugging || checking) printf("numpoints is: %ld\n",numpoints);
/*
    In the next few statements, we allocate "parallel arrays" which are
    being used to implement records containing fields for as many
    coordinates as the embedding dimension requires, for a "next_after"
    pointer field to be used to link the records into lists, and for a
    "marker" field to indicate whether the element is being used to mark the
    end of a pass in the radix sort.
*/

/* allocate memory to point to a coordinate array for each coordinate --
   embedding dimension determines how many coordinates.  */
        if ( (data_array = (ulong **) malloc
	              (embed_dim*sizeof(ulong *))) == NULL)
	{
	  printf("Memory allocation failure: \"data_array\".\n");
	  exit(-1);
	}
/* Here we allocate the storage for the coordinates of data points, plus
   extra room for 2 queue markers needed in the radix sort.  */   
	for (i=0;i<embed_dim;i++)
          if ((data_array[i] = (ulong *) malloc ((numpoints+2)*sizul)) == NULL)
	    {
		printf("Memory allocation failure: \"data_array[%d]\").\n",i);
		exit(-1);
            }
/* Here we allocate the storage for the pointers.  These are implemented as
   long integers.  The values of these pointers will be used as indexes into
   the arrays being declared here.  Note that here we also allocate room for
   two markers. */   
          if((next_after=(long *)malloc((numpoints+2)*sizeof(long))) == NULL)
	    {
		printf("Memory allocation failure: \"next_after\" array).\n");
		exit(-1);
            }
	    
/* Here we allocate storage for a tag field that tells whether a record is a
   marker. */
	  if ( (marker = (int *) malloc
	                      ( (numpoints+2)*sizeof(int) ) ) == NULL)
	    {
		printf("Memory allocation failure: \"marker\" array).\n");
		exit(-1);
            }
/* Here we allocate storage for a word that describes the bit changes
   between two different unsigned long int's.  We have one of these words
   for each of the embedding dimensions.  These will come in handy when the
   sweep is done that gets the box counts. */
	  if ( (diff_test = (ulong *) malloc
	                      ((embed_dim)*sizeof(ulong)) ) == NULL)
	    {
		printf("Memory allocation failure: \"diff_test\" array).\n");
		exit(-1);
            }

/* Initialize the queues to an empty state.  We load the data directly into
   these queues, and use them afterwards to do the radix sort and the sweep.
*/
	create_Q(&Q[0]); create_Q(&Q[1]);
	if (debugging || checking) printf ("Just created queues.\n");
        if (debugging) print_Q(&Q[0]);
        if (debugging) print_Q(&Q[1]);

/* Initialize the avail list to hold all the records implemented by the
   parallel arrays.
*/
	create_Av(&avail)   ;
        if (debugging || checking) printf ("Just created avail.\n");
        if (debugging) print_Av(avail);

/* Place a marker in each queue. Markers will be in front of each queue
   after all the data has been loaded. */
        for (i=0;i<2;i++)
	{
	   pop_Av(&avail, &pointer);
	   marker[pointer] = 1;
   	   en_Q(&(Q[i]),pointer);
	}
/* Place all the data points in the appropriate queue. */
   	for(i=0;i<numpoints;i++)
	{
           pop_Av(&avail, &pointer);   /* get next record */
	   marker[pointer] = 0;        /* it is not a marker */
	   for (j=0; j<embed_dim;j++)  
	   {  fscanf(infile,"%lf",&buf);  /* put scaled data in array */
	      if (debugging) printf ("Just read %lf\n",buf);
              data_array[j][pointer] =			
	                (ulong)( (buf-min) * (maxdiam/(max-min)) );	      
	   }
/* Get started on the radix sort by putting the data in the queue
   corresponding to the least significant bit of the last coordinate. */
   	   if (IS_A_ONE(data_array[embed_dim-1][pointer],0))
	    en_Q(&(Q[1]),pointer);       /* last coord ends in "1" */
	   else en_Q(&(Q[0]),pointer);   /* last coord ends in "0" */
	}

  /* close the input file */
fclose(infile);
if (debugging || checking) printf ("Just through loading data.\n");
if (debugging) print_Av(avail);
if (debugging) print_Q(&Q[0]);
if (debugging) print_Q(&Q[1]); 

   /* radix sort queues */
radixsort();

  /* sweep data_array */
sweep(box_count_array);

      /* write box counts to standard output. */
for(i=0;i<=numbits;i++)
printf("box_count_array[%d] =\t%lu\n",i,box_count_array[i]);

    /* find and print Fractal Dimension */
fd = fracdim(box_count_array,numpoints);
printf("\n\nFractal Dimension (Capacity Method): %f\n",fd);

}
