/* BEGIN NOTICE

This software was written by Peter DiFalco and John Sarraille.  The
algorithm employed was inspired by the paper entitled "A Fast
Algorithm To Determine Fractal Dimensions By Box Counting", which was
written by Liebovitch and Toth, and which appeared in the journal
"Physics Letters A", volume 141, pp 386-390, (1989).

This software is not warranteed: use at your own risk.  Unlimited
permission to distribute this software is granted, so long as this
notice is included in all copies made.

Sept 2, 1991

END NOTICE */


#include <stdio.h>
#include "fd.h"
#include <math.h>

int get_e_dim();
void max_min();
void rad_pass();
void radixsort();
void sweep();
float fracdim();

int get_e_dim(userinputfile)
	char *userinputfile;
{
   extern FILE  *infile;
   int c, count ;   double temp ;
   if (debugging || checking) printf("Now inside get_e_dim.\n");
   if (debugging || checking) printf("About to open input file.\n");   
   if((infile = fopen(userinputfile,"r")) == NULL)  /* open input file */
     {
       printf("Cannot open input file, %s\n",userinputfile);
       exit(-1);
     }
      /* get past "numpoints", and the first number on the first line of
         actual data. */
   fscanf (infile, "%f%f", &temp, &temp);
   count = 1 ;        /* One number on this line, so far. */
   do      /*  Read and count the rest of the numbers on this line. */
    {      
      do     c=getc(infile) ;          /* Skip over same-line white space */
      while  ( (c=='\t') || (c==' ')  ) ;
      if   ( (c != '\n') && (c != '\r') && (c !='\f') )
       {                       /* if we don't seem to be on a new line */
        count = count + 1 ;    /* then increment # of numbers on this line */
        do   c=getc(infile) ;  /* and read past that next number */
        while  ( (c != '\n') && (c != '\r') && (c !='\f')
                 && (c !='\t') && (c != ' ') );
       }
    }  while ( (c != '\n') && (c != '\r') && (c !='\f') ) ;
   fclose (infile);
   return (count);
}

void max_min(userinputfile,pmax, pmin)
	char *userinputfile;
	double *pmax, *pmin ;
{
        extern int   embed_dim;      /* How many coordinates points have. */
	FILE	*infile;	/* input file */
	double	temp;
	ulong	i, numToRead;		/* control variable */
	ulong	numpoints;	/* number of data points in input file */
	/* open input file */
	if((infile = fopen(userinputfile,"r")) == NULL)
	{
		printf("Cannot open input file, %s\n",userinputfile);
		exit(-1);
	}
	/* get numpoints */
	if ((fscanf(infile,"%lu",&numpoints)) == 0)
	  {
	     printf ("Format of input file, %s, ", userinputfile);
	     printf ("is incorrect -- please check.\n");
	     exit(-1);
	  }
	if (debugging || checking) printf ("Now control is in max_min.\n");
	if (debugging || checking) printf ("Numpoints is %lu.\n", numpoints);
	/* find maximum and minimum data values in the input file.
	   These are needed so that the data can be scaled to be a set of
	   non-negative integers between 0 and maxdiam, where maxdiam will
	   be the largest integer value expressible as an element of the
	   data type "unsigned long int". */
	if (debugging || checking) printf ("Starting to read data.\n");
	fscanf(infile,"%lf",&temp);
	if (debugging || checking) printf ("First datum is %lf.\n", temp);
	*pmin=*pmax=temp;
	if (debugging) printf ("Pmin is: %lf, Pmax is: %lf\n", *pmin, *pmax);
	numToRead = ((ulong) (embed_dim)) * numpoints;
	if (debugging || checking) printf ("numToRead is: %lu\n", numToRead);
	for(i=1;i<numToRead;i++)
	{
		fscanf(infile,"%lf",&temp);
		if (debugging) printf ("Next datum read is: %lf\n", temp);
		if(temp > *pmax)
			*pmax = temp;
		else
			if(temp < *pmin)
				*pmin = temp;
 if (debugging) printf ("Pmin is now: %lf, Pmax is now: %lf\n", *pmin, *pmax);
	}
	/* check to see if the maximum equals the minimum -- this is a
	   degenerate case -- or it might mean that the input file is
}	   faulty. */
	if(*pmax == *pmin)
        {
	  printf ("The input file, %s, is confusing!\n\n", userinputfile);
	  printf ("Either all the points in %s have the same ", userinputfile);
	  printf ("coordinates,\n");
	  printf ("(THE FRACTAL DIMENSION IS ZERO IN THIS CASE.)\n\n");
	  printf ("or %s is simply of the wrong form for an\n",userinputfile);
	  printf ("input file to this program -- please check.\n");
	  exit(-1);
	}
	/*  close the input file */
	fclose(infile);
}

/*
   This procedure performs one pass of the radix sort.  It assumes that the
   queues each have a marker in front at the time of the call, and this is
   the condition the procedure LEAVES the queues in when it terminates.
*/   
void rad_pass(coord, bitpos)
	int coord,   /*  The coordinate we are doing this pass on */
	    bitpos;  /*  The bit position we are doing this pass on */
{  extern int  *marker;           /*  To mark queues during radix sort */
   extern QHeader Q[2];       /* array of two queues for the radix sort */
   extern ulong **data_array;	/* initial array of data points */   
   int queue_num, index ;
    /*  Move the markers to the rear of the queues */
 if (debugging)
    printf ("Starting pass on bit %d of coord %d.\n",bitpos,coord);    
 if (debugging) printf ("Moving markers to the rear of queues.\n");
 for (queue_num=0;queue_num<2;queue_num++)
   {  de_Q(&(Q[queue_num]),&index); en_Q(&(Q[queue_num]),index);  }
 if (debugging) print_Q(&Q[0]);
 if (debugging) print_Q(&Q[1]);

    /*  Move all non-marker elements to the appropriate queue. */
 if (debugging) printf ("Starting main part of pass.\n");    
 for (queue_num=0;queue_num<2;queue_num++)
   {    /* Peek first to see if Qfront is a marker -- this violates
	   information hiding, and would be "cleaner" if done with a
	   procedure in the queue package.  */
     while ( marker[Q[queue_num].Qfront] == 0 ) 
       {     /*  Peeking again! */
         if (IS_A_ONE(data_array[coord][Q[queue_num].Qfront],bitpos))
 	       /* Directly transfer from the source to target queue */
          transf_Q( &(Q[queue_num]),&(Q[1]) );
	 else transf_Q( &(Q[queue_num]),&(Q[0]) ) ;
	if (debugging) printf("A queue transfer is done.  \n");
        if (debugging) print_Q(&Q[0]);
        if (debugging) print_Q(&Q[1]);
       }  }  }

/*
This sort is to be used directly after fddriver does the initial loading of
the data into the queues.  It leaves the data essentially sorted, with all
the data whose x-coord starts with 0 in Q[0], in order, and all the data
whose x-coord starts with 1 in Q[1], in order.  Thus the sweep that comes
next must traverse Q[0], and then Q[1].
*/
void radixsort()
{
  extern QHeader Q[2];       /* array of two queues for the radix sort */
  extern int   embed_dim;      /* How many coordinates points have. */
  extern long int  avail;             /*  avail list pointer */  
  int  bitpos, coord, queue_num;
  long int index;

    /* Finish up on the zero-bit -- loading data took care of first pass. */
  for (coord=embed_dim-2;coord>=0;coord--) rad_pass(coord,0);
	
      /* Now sort on the rest of the bits. */
  for (bitpos=1;bitpos<numbits;bitpos++)
      for (coord=embed_dim-1;coord>=0;coord--) rad_pass(coord,bitpos);

      /*  Lastly, get the markers out of the queues. */
  for (queue_num=0;queue_num<2;queue_num++)
    {
      de_Q(&(Q[queue_num]),&index);
      push_Av(&avail,index);
    } }

void sweep(count_array)
	ulong	count_array[numbits+1];	/* counter array */
{ extern int  embed_dim;    /* How many coordinates points have. */
  extern ulong **data_array;	/* initial array of data points */   
  extern QHeader Q[2];       /* array of two queues for the radix sort */
  extern ulong  *diff_test;   /* XOR's of pairs of coordinates */
  extern long int  *next_after;    /*  array for pointers to data_array. */  
  int        bitpos, countpos, coord, queue_num, found;
  long int   current, previous;

    /* Get a pointer to the first data value */
  if    (Q[0].Qfront != -1) previous = Q[0].Qfront;
  else  previous = Q[1].Qfront;

     /* init count_array -- first point counts as one box of every size.*/
  for(countpos=0;countpos<=numbits;countpos++)	count_array[countpos]=1;

     /* Now sweep the data, comparing new points with the previous, to see
	when new boxes at each level are entered. */
  for (queue_num=0;queue_num<2;queue_num++)
    { current = Q[queue_num].Qfront;
      while (current != -1)
        {  found = 0 ;  bitpos=numbits-1;
           for (coord=embed_dim-1;coord>=0;coord--)
             diff_test[coord] = data_array[coord][previous]^
                                      data_array[coord][current];
           do { coord = embed_dim - 1;
	        do {  if ( IS_A_ONE(diff_test[coord],bitpos) )
	                {  for (countpos=bitpos;countpos>=0;countpos--)
		           count_array[countpos]++ ;
		           found = 1;
		        }
                      else coord-- ;
	           }
	        while ( (found == 0) && (coord > -1) );
	        bitpos-- ;
              }
           while ( (found == 0) && (bitpos > -1) );
           previous = current; current = next_after[current];
        }
    }
}

float fracdim(box_count_array,numpoints)
	ulong	box_count_array[numbits+1],
		numpoints;
{
	int	i,j=0,			/* control variables */
	        cut_off=1,                /* saturation constant */
		mark=0;			/* marks the index for > N/cut_off */
	float	tempfd;		/* The Fractal Dimension to be returned */
	double	xmean=0,ymean=0,	/* mean value for x's and y's */
		temp1=0,temp2=0,	/* variables */
		fdpoints[2][numbits-1];

 /* Calculate cut_off */
 for (i=1;i<=embed_dim;i++) cut_off = cut_off * 2;
 cut_off++;
/*
    Mark greatest index where count > box_count_array[0]/cut_off.
    Counts at lesser indexes will not be used in the estimate of
    fractal dimension -- distortion due to saturation is the concern.

    Note that box_count_array[0] is the number of boxes of size 1
    (the smallest size) that contain a point of the set.  For all
    practical purposes, box_count_array[0] will equal the number of
    DISTINCT points in the input file, because these boxes are
    REALLY SMALL compared to the size of the biggest box (this size
    is 4 billion if an unsigned long int is 32 bits to the platform
    compiler.  The points are scaled by the program so that the set
    is too "large" to fit in the next smallest box-size, and so that
    "1" is the smallest difference in value that can be resolved.)
    One box, in effect, covers only a single point of the input set
    because the program can't resolve points with a smaller
    difference.

    It would be a bad idea to use numpoints/cut_off as the limit
    because in cases where there were many duplicate points, we
    would seriously over-estimate the number of distinct points, and
    thus use saturated data to base the estimate of fractal
    dimension upon.  When testing the program with random data
    sampled with replacement, this could throw the results way off.
    (This happened to us, and it took us a while to figure out why.
    Afterwards, we stopped using numpoints/cut_off, and changed to
    box_count_array[0]/cut_off.)
*/
	for(i=0;i<numbits;i++)
	{
	   if(box_count_array[i] > box_count_array[0]/cut_off)
	     mark=i;
	}
        if (mark <= numbits -2)
        {
        printf ("\nThe smallest box size usable in the ");
	printf ("estimate has index:  %d\n", mark);
        printf ("The largest box size used in the estimate ");
	printf ("is at index: %d\n\n", numbits-2) ;
        printf ("Data whose slope will be used to calculate ");
	printf ("Fractal Dimension:\n\n");
	printf("   log[1/e]         log[N(e)]        log[N(e)-N(2e)]\n\n");
        }

	for(i=numbits-2;i>=mark;i--)  
	{
		/* set i'th "x" value to -i -- this is the log (base 2) of
    		   the inverse of the diameter of the i'th size of box.  */
		fdpoints[0][i] = -i;

		/* print the value of x */
                printf ("%8d",-i);

                /* update calculation of the mean x-value. */
		xmean = xmean - i;

		/* set the i'th "y" value to log (base 2) of number    */
                /* of boxes of side 2^i required to cover the set */
		fdpoints[1][i] = log((double) (box_count_array[i]))/log(2.0);

		/* print the value of y */
                printf ("          %12f",fdpoints[1][i]);

		/* print the value of y(i)-y(i-1) */
                if (i<numbits-2)
                printf ("       %12f",fdpoints[1][i]-fdpoints[1][i+1]);
                printf ("\n");

                /* update calculation of the mean y-value. */
                ymean = ymean + log((double) (box_count_array[i]))/log(2.0);
		j++;
	}
	xmean = xmean/((double) j);
	ymean = ymean/((double) j);

	for(i=numbits-2;i>=mark;i--)
	{
            temp1 = temp1 + ((fdpoints[0][i]-xmean)*(fdpoints[1][i]-ymean));
	    temp2 = temp2 + ((fdpoints[0][i]-xmean)*(fdpoints[0][i]-xmean));
	}

	if(temp2 == 0)
	{
	   printf("\nInsufficient data. More points are needed.\n");
           printf("Cannot assign a Fractal Dimension to this set.\n");
           printf("Examine the data above to make your own conjecture.\n");
           exit(-1) ;   
	}
	else
	{	/* return the Fractal Dimension */
		tempfd = (float) (temp1/temp2);
		return(tempfd);
	}
}
